/*
 * Author: Ulf Lilleengen <lulf@kerneled.org>
 *
 * Copyright(C) 2004-2005 Ulf Lilleengen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <pthread.h>

#include <config.h>
#include <wnewsgtk.h>
#include <headline.h>
#include <socket.h>
#include <wstring.h>
#include <rss.h>

#ifdef __linux__
#define strlcat strncat
#define strlcpy strncpy
#endif

#define PADVAL 4
#define MAXLN 1024

GtkWidget *window_main;
GtkWidget *desc_entry;
GtkWidget *title_entry;
GtkWidget *address_entry;
GtkTreeStore	*store;
GtkWidget *dialog;
GtkWidget *list;
GtkTreeIter iter_global;

pthread_mutex_t list_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t store_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t iter_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t resc_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_t news_tr;

GtkWidget *create_widgets(gboolean, gint);
int fill_cbox(GtkComboBox *);
int resc_remove(GtkWidget *, gpointer);
void dialog_add(GtkWidget *, gpointer);
int resc_add(GtkWidget *, gpointer);
GtkTreeModel * create_model(void);
GtkWidget * create_view(void);
int addToList(GtkWidget *, gpointer);
void *add_to_list_thread(void *);
void expand_collapse_row(GtkTreeView *, GtkTreePath *, GtkTreeViewColumn *, gpointer);
int rightClickMenu(GtkWidget *, GdkEventButton *, gpointer);
int openInBrowser(GtkMenuItem *, gpointer);
int copy_to_clipboard(GtkWidget *, gpointer);
int copy_url_to_clipboard(GtkWidget *, gpointer);

/*
 * Starts the gtk+ version of wnews. Initializes window, load resources
 * run main window loop.
 */
void startGtk(void)
{
	GtkWidget *box;
	window_main = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_resize(GTK_WINDOW(window_main), 500, 600);
	gtk_window_set_default_size(GTK_WINDOW(window_main), 500, 600);
	gtk_window_set_resizable (GTK_WINDOW (window_main), FALSE);

	box = create_widgets(0, PADVAL);
	gtk_container_add(GTK_CONTAINER(window_main), box);
	gtk_widget_show(box);
	g_signal_connect (G_OBJECT (window_main), "destroy", G_CALLBACK (destroy), NULL);
	gtk_widget_show(window_main);
	gtk_main();
}

/*
 * Creates main widgets as buttons, combobox and listview. Packs
 * the widgets in horizontal and vertical boxes, and connects
 * signals to various functions.
 */
GtkWidget *
create_widgets(gboolean homogeneous, gint spacing)
{
	GtkWidget *box;
	GtkWidget *buttonbox;
	GtkWidget *button_add;
	GtkWidget *button_del;
	GtkWidget *button_go;
	GtkWidget *scrollist;
	GtkWidget *combobox_sitelist;
	GtkWidget *separator;
	GtkWidget *listbox;

	box = gtk_vbox_new(homogeneous, spacing);
	
	buttonbox = gtk_hbox_new(0, PADVAL);
	
	combobox_sitelist = gtk_combo_box_new_text();
	gtk_box_pack_start (GTK_BOX(buttonbox), combobox_sitelist, TRUE, TRUE, PADVAL);
	gtk_widget_show(combobox_sitelist);

	separator = gtk_hseparator_new();
	gtk_box_pack_start (GTK_BOX(buttonbox), separator, FALSE, TRUE, 10);
	gtk_widget_show(separator);
	
	button_add = gtk_button_new_with_label("Add");
	gtk_box_pack_start (GTK_BOX(buttonbox), button_add, FALSE, FALSE, PADVAL);
	gtk_widget_show(button_add);

	button_del = gtk_button_new_with_label("Delete");
	gtk_box_pack_start (GTK_BOX(buttonbox), button_del, FALSE, FALSE, PADVAL);
	gtk_widget_show(button_del);

	button_go = gtk_button_new_with_label("Go");
	gtk_box_pack_start (GTK_BOX(buttonbox), button_go, FALSE, FALSE, PADVAL);
	gtk_widget_show(button_go);

	gtk_box_pack_start (GTK_BOX(box), buttonbox, FALSE, FALSE, PADVAL);
	gtk_widget_show(buttonbox);
	
	listbox = gtk_hbox_new(0, PADVAL);
	scrollist = gtk_scrolled_window_new(NULL, NULL);
	list = create_view();
	gtk_widget_set_size_request(list, -1, 300);
	gtk_container_add(GTK_CONTAINER(scrollist), list); /* Add scrollbar to treeview */
	gtk_widget_show(list);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollist), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start (GTK_BOX(listbox), scrollist, TRUE, TRUE, PADVAL);
	gtk_widget_show(scrollist);
	gtk_box_pack_start (GTK_BOX(box), listbox, TRUE, TRUE, PADVAL);
	gtk_widget_show(listbox);
	
	fill_cbox(GTK_COMBO_BOX(combobox_sitelist));

	g_signal_connect_swapped(G_OBJECT(button_go), "clicked", G_CALLBACK(addToList), combobox_sitelist);
	g_signal_connect_swapped(G_OBJECT(button_del), "clicked", G_CALLBACK(resc_remove), combobox_sitelist);
	g_signal_connect_swapped(G_OBJECT(button_add), "clicked", G_CALLBACK(dialog_add), combobox_sitelist);
	g_signal_connect(G_OBJECT(list), "row-activated", G_CALLBACK(expand_collapse_row), NULL);
	g_signal_connect_swapped(G_OBJECT(list), "button_press_event", G_CALLBACK(rightClickMenu), window_main);
	return box;
}

/*
 * Fills the combobox with data from the resources file. Appends
 * according the resources file, and sets the first resource active.
 */
int fill_cbox(GtkComboBox *combobox_sitelist)
{
	char *entrytext;
	int i = 0, bytes = 0;
	struct resource *grl, **grarray;

	numr = 0;
	if ((grl = rconfig()) == NULL) {
		fprintf(stderr, "Error loading resources configuration file\n");
		return -1;
	}
	grarray = rhash(grl, numr - 1);
	for (i = 0; i < numr; i++) {
		bytes = (strlen(grarray[i]->title)+strlen(grarray[i]->desc)+4);
		if ((entrytext = malloc(bytes)) == NULL) {
			fprintf(stderr, "Error allocating memory for the combobox entry\n");
			rescfree(grarray, numr);
			return -1;
		}
		strlcpy(entrytext, grarray[i]->title, bytes);
		strlcat(entrytext, " - ", bytes);
		strlcat(entrytext, grarray[i]->desc, bytes);
		gtk_combo_box_append_text((GtkComboBox*)combobox_sitelist, entrytext);
		free(entrytext);
		bytes = 0;
		entrytext = 0;
	}
	rescfree(grarray, numr);
	gtk_combo_box_set_active(combobox_sitelist, 0);
}

/*
 * Create a GtkTreeView, holding one column with a string.
 */
GtkWidget *
create_view(void)
{
	GtkWidget			*treeview;
	GtkTreeModel		*model;
	GtkCellRenderer	*rend;
	GtkTreeViewColumn *col1;

	treeview = gtk_tree_view_new();
	
	/* Create column 1 */
	rend = gtk_cell_renderer_text_new();
	col1 = gtk_tree_view_column_new_with_attributes(NULL, rend, "text", 0, "text", 1, NULL);
	gtk_tree_view_column_set_sizing(col1, GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_column_set_min_width(col1, 480);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col1);
	
	model = create_model();
	gtk_tree_view_set_model(GTK_TREE_VIEW(treeview), model);
	g_object_unref(model);
	return treeview;
}

/*
 * Create a GtkTreeModel with a column displaying a string.
 */
GtkTreeModel *
create_model(void)
{
	GtkTreeIter		iter;

	/* Create tree store with 1 column */
	store = gtk_tree_store_new (2, G_TYPE_STRING, G_TYPE_STRING);

	return GTK_TREE_MODEL (store);
}
	
/*
 * Add the currently selected item in the combobox to the
 * tree list. Search the treemodel for matching headers and
 * fetch news according to the headline structure's address field.
 */
int addToList(GtkWidget *widget, gpointer callback_data)
{
	pthread_create(&news_tr, NULL, add_to_list_thread, (void *)widget);
	return 0;
}

void *
add_to_list_thread(void *a)
{
	GtkWidget *msgdlg;
	GtkComboBox *combobox_sitelist = GTK_COMBO_BOX(a);
	GtkTreeModel *mod;
	struct content **gcarray, *cont;
	char *snews = NULL, *headline, *desc;
	struct resource *grl, **grarray;
	GtkTreeIter iter, iter_child, iter_combo, iter_grandchild;
	gchar *site, *listdata;
	gboolean valid;
	int contsize, i, num;

	/* Decide the text width according to column size. Lock treeview first */
	pthread_mutex_lock(&list_lock);
	contsize = (gtk_tree_view_column_get_width(gtk_tree_view_get_column(GTK_TREE_VIEW(list), 0)))/6-25;
	pthread_mutex_unlock(&list_lock);

	pthread_mutex_lock(&store_lock);
	mod = gtk_combo_box_get_model(combobox_sitelist);
	gtk_combo_box_get_active_iter(combobox_sitelist, &iter_combo);
	gtk_tree_model_get(mod, &iter_combo, 0, &site, -1);
	valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(store), &iter);

	while (valid)
	{
		gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, 0, &listdata, -1);
		if (strcmp(site, listdata) == 0) /* Already exists in view*/
			gtk_tree_store_remove(store, &iter);

		g_free(listdata);
		valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter);
	}
	gtk_tree_store_prepend(store, &iter, NULL);
	gtk_tree_store_set(store, &iter, 0, site, -1);
	pthread_mutex_unlock(&store_lock);

	pthread_mutex_lock(&resc_lock);
	numr = 0;
	if ((grl = rconfig()) == NULL) {
		fprintf(stderr, "Error loading resources configuration file\n");
		return NULL;
	}
	grarray = rhash(grl, numr - 1);
	for (i = 0; i < numr; i++)
		if (strstr(site, grarray[i]->title) != NULL)
			break;
	pthread_mutex_unlock(&resc_lock);
	g_free(site);
	
	pthread_mutex_lock(&iter_lock);
	iter_global = iter;
	pthread_mutex_unlock(&iter_lock);
	if ((snews = getpages(grarray[i]->address, proxy)) == NULL) {
		werror("Error getting news");
		return NULL;
	}
	cont = getrss(snews);
	pthread_mutex_lock(&store_lock);
	gcarray = chash(cont, numc - 1);
	for (i = 0; i < numc; i++) {
		gtk_tree_store_append(store, &iter_child, &iter);
		headline = strtagstrip(gcarray[i]->headline);
		gtk_tree_store_set(store, &iter_child, 0, headline, 1, gcarray[i]->link, -1);
		free(headline);
		if(strcmp(gcarray[i]->desc, "") != 0) {
			gtk_tree_store_append(store, &iter_grandchild, &iter_child);
			desc = strtagstrip(gcarray[i]->desc);
			free(gcarray[i]->desc);
			gcarray[i]->desc = strcol(desc, contsize);
			free(desc);
			gtk_tree_store_set(store, &iter_grandchild, 0, gcarray[i]->desc, 1, gcarray[i]->link, -1);
		}
	}
	contfree(gcarray, numc);
	rescfree(grarray, numr);
	pthread_mutex_unlock(&store_lock);
	free(snews);
}

/*
 * View the text to the headline which was clicked in the text-browser
 */
void
expand_collapse_row(GtkTreeView *tv, GtkTreePath *path, GtkTreeViewColumn *col, gpointer callback_data)
{
	if( gtk_tree_view_row_expanded(tv, path) )
		gtk_tree_view_collapse_row(tv, path);
	else
		gtk_tree_view_expand_row(tv, path, FALSE);
}

/*
 * Handle right click event on treelist, displaying a popup menu
 * presenting choices like copy, and open in browser.
 */
int
rightClickMenu(GtkWidget *widget, GdkEventButton *event, gpointer callback_data)
{
	gint depth;
	GtkTreeModel *mod;
	GtkTreePath	*path;
	GtkTreeViewColumn *col;
	GtkTreeIter iter;
	GtkWidget *menuitem1, *menuitem2, *menuitem3;
	GtkWidget *menu;
	gchar *buf, *buf_addr;
	gtk_tree_view_get_cursor((GtkTreeView*)callback_data, &path, &col);
	mod = gtk_tree_view_get_model((GtkTreeView *)callback_data);

	if( event->type == GDK_BUTTON_PRESS) {
		if( event->button == 3 ) {
			menu = gtk_menu_new();
			switch ( (depth = gtk_tree_path_get_depth(path)))
			{
				case 1:
					gtk_tree_path_free(path);
					return 0;
				break;
				case 2: case 3:
					/* Adding item */
					menuitem1 = gtk_menu_item_new_with_label("Open in browser");
					menuitem2 = gtk_menu_item_new_with_label("Copy");
					menuitem3 = gtk_menu_item_new_with_label("Copy URL");
				break;
				default:
					gtk_tree_path_free(path);
					return 0;
			}
			gtk_tree_model_get_iter(mod, &iter, path);
			gtk_tree_model_get(mod, &iter, 0, &buf, 1, &buf_addr, -1);
			gtk_menu_shell_append( GTK_MENU_SHELL(menu), menuitem1);
			gtk_menu_shell_append( GTK_MENU_SHELL(menu), menuitem2);
			gtk_menu_shell_append( GTK_MENU_SHELL(menu), menuitem3);
			gtk_widget_show(menuitem1);
			gtk_widget_show(menuitem2);
			gtk_widget_show(menuitem3);
			switch(depth) {
			case 2: case 3:
				
				g_signal_connect(G_OBJECT(menuitem1), "activate", G_CALLBACK(openInBrowser), buf_addr);
				g_signal_connect(G_OBJECT(menuitem2), "activate", G_CALLBACK(copy_to_clipboard), buf);
				g_signal_connect(G_OBJECT(menuitem3), "activate", G_CALLBACK(copy_url_to_clipboard), buf_addr);
			
			break;
			}
			gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
		}
	}
	gtk_tree_path_free(path);
	return 0;
}

/*
 * Open an url in a webbrowser.
 */
int
openInBrowser(GtkMenuItem *widget, gpointer callback_data)
{
	char *argv = (char *)callback_data;
	char *command;
	
	if( browser == NULL ) {
		dialog = gtk_message_dialog_new (GTK_WINDOW(window_main), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_CLOSE, "You need to set the browser variable in \"~/.wnews/config\". Ex. \"BROWSER firefox %%s\"");
		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);
		return 0;
	}
	if ((command = malloc(strlen(argv) + strlen(browser) + 1)) == NULL) {
		fprintf(stderr, "Error allocating memory for browser command\n");
		return -1;
	}
	
	sprintf(command, browser, argv);
	system(command);
	free(command);
	free(argv); /*XXX not sure if this is nessecary, but gives no warnings */
	return 0;
}

/*
 * Destroy the gtk+ window.
 */
void
destroy(GtkWidget *widget, gpointer data)
{
	if (news_tr)
		pthread_cancel(news_tr);
	gtk_main_quit();
}

/*
 * Create a dialog for adding a new resource.
 */
void
dialog_add(GtkWidget *widget, gpointer data)
{
	GtkWidget *dialog, *label, *dialogbox, *add_button, *close_button, *buttonbox;
	dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	dialogbox = gtk_vbox_new(0, 1);
	gtk_container_add(GTK_CONTAINER(dialog), dialogbox);
	label = gtk_label_new("Add a new site to the roll-down list");
	title_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(title_entry), "Enter title");

	desc_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(desc_entry), "Enter description of site");
	
	address_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(address_entry), "Enter address to the site");
	
	add_button = gtk_button_new_with_label("Add");
	close_button = gtk_button_new_with_label("Close");

	gtk_box_pack_start (GTK_BOX (dialogbox), label, FALSE, TRUE, 2);
	gtk_misc_set_padding(GTK_MISC(label), 5, 5);

	gtk_box_pack_start (GTK_BOX (dialogbox), title_entry, FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX (dialogbox), desc_entry, FALSE, FALSE, 2);
	gtk_box_pack_start (GTK_BOX (dialogbox), address_entry, FALSE, FALSE, 2);
	
	buttonbox = gtk_hbox_new(0, 2);
	gtk_box_pack_start (GTK_BOX (buttonbox), close_button, TRUE, FALSE, 1);
	gtk_box_pack_start (GTK_BOX (buttonbox), add_button, TRUE, FALSE, 1);
	
	gtk_box_pack_start (GTK_BOX (dialogbox), buttonbox, FALSE, FALSE, 2);
	g_signal_connect_swapped(G_OBJECT(add_button), "clicked", G_CALLBACK(resc_add), widget);
	g_signal_connect_swapped(G_OBJECT(close_button), "clicked", G_CALLBACK(gtk_widget_destroy), dialog);
	gtk_widget_show(label);
	gtk_widget_show(title_entry);
	gtk_widget_show(desc_entry);
	gtk_widget_show(close_button);
	gtk_widget_show(address_entry);
	gtk_widget_show(add_button);
	gtk_widget_show(buttonbox);
	gtk_widget_show(dialogbox);
	gtk_widget_show(dialog);
}

/*
 * Add new resource specified by the results of the add dialog.
 * Add the resource to the resources file, as well as the combobox.
 */
int
resc_add(GtkWidget *widget, gpointer data)
{
	FILE *fp;
	char *fname, *entrytext;
	int bytes = 0;
	GtkComboBox *cbox = GTK_COMBO_BOX(widget);
	if (!(fname = malloc(LMAX+1))) {
		fprintf(stderr, "Could not allocate memory for file name buffer\n");
		return -1;
	}

	sprintf(fname, "%s%s", me->pw_dir, RSOURC);
	if( (fp = fopen(fname, "a")) == NULL) {
		fprintf(stderr, "Error, can't open resources file\n");
		return -1;
	}
		
	fprintf(fp, "%s|%s|%s\n", gtk_entry_get_text(GTK_ENTRY(title_entry)), gtk_entry_get_text(GTK_ENTRY(desc_entry)), gtk_entry_get_text(GTK_ENTRY(address_entry)));
	fclose(fp);
	
	bytes = (strlen(gtk_entry_get_text(GTK_ENTRY(title_entry)))+strlen(gtk_entry_get_text(GTK_ENTRY(desc_entry)))+4);
	if ( (entrytext = malloc(bytes)) == NULL) {
		fprintf(stderr, "Error allocating memory for combobox entry");
		return -1;
	}
	strlcpy(entrytext, gtk_entry_get_text(GTK_ENTRY(title_entry)), bytes);
	strlcat(entrytext, " - ", bytes);
	strlcat(entrytext, gtk_entry_get_text(GTK_ENTRY(desc_entry)), bytes);
	gtk_combo_box_append_text(cbox, entrytext);
	free(entrytext);
	free(fname);
	return 0;
}

/*
 * Removes currently selected entry from resources file, as well
 * as the combobox and the treeview widgets.
 */
int
resc_remove(GtkWidget *widget, gpointer data)
{
	FILE *fp_old, *fp_new;
	char *junk, *junk2, *line, *headl_news, *parse;
	gint num = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
	char *text = gtk_combo_box_get_active_text(GTK_COMBO_BOX(widget));
	gchar *listdata;
	gboolean valid;
	GtkTreeIter iter;
	GtkTreePath *path;
	
	if (text == NULL)
		return -2;
	gtk_combo_box_remove_text(GTK_COMBO_BOX(widget), num);
	gtk_combo_box_set_active(GTK_COMBO_BOX(widget), 0);

	if ((line = malloc(LMAX+1)) == NULL ) {
		fprintf(stderr, "Could not allocate memory for line buffer\n");
		return -1;
	}
	if ((junk = malloc(LMAX+strlen(RSOURC)+1)) == NULL) {
		fprintf(stderr, "Could not allocate memory for junk buffer\n");
		return -1;
	}
	if ((junk2 = malloc(LMAX+strlen(RSOURC)+5)) == NULL) {
		fprintf(stderr, "Could not allocate memory for junk2 buffer\n");
		return -1;
	}
	sprintf(junk, "%s%s", (char *)me->pw_dir, (char *)RSOURC);
	if ((fp_old = fopen(junk, "r")) == NULL) {
		fprintf(stderr, "Could not open resources file\n");
		return -1;
	}
	sprintf(junk2, "%s%s_new", (char *)me->pw_dir, (char *)RSOURC);
	if ((fp_new = fopen(junk2, "a")) == NULL) {
		fprintf(stderr, "Could not open file\n");
		fclose(fp_old);
		return -1;
	}
	while (fgets(line, LMAX, fp_old) != NULL) {
		parse = strdup(line);
		if (parse[0] == '#') {
			fprintf(fp_new, line);
			free(parse);
			continue;
		}
		if ( (headl_news = strtok(parse, "|")) != NULL) {
			if ((strstr(text, headl_news)) == NULL)
				fprintf(fp_new, line);
		}
		free(parse);
	}
	remove(junk);
	rename(junk2, junk);
	fclose(fp_old);
	fclose(fp_new);
	
	pthread_mutex_lock(&store_lock);
	valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(store), &iter);
	while (valid) 
	{
		gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, 0, &listdata, -1);

		if (strcmp(text, listdata) == 0) {
			gtk_tree_store_remove(store, &iter);
			gtk_tree_view_set_model(GTK_TREE_VIEW(list), GTK_TREE_MODEL(store));
			g_free(listdata);
			break;
		}
		g_free(listdata);
		valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter);
	}
	pthread_mutex_unlock(&store_lock);
	free(junk2);
	free(junk);
	free(line);
	return 0;
}

/*
 * Copy the current marked content into a clipboard
 */
int
copy_to_clipboard(GtkWidget *widget, gpointer data)
{
	GtkClipboard *cb = gtk_clipboard_get(GDK_NONE);
	gtk_clipboard_set_text(cb, (char *)data, strlen((char *)data));
	return 0;
}

/*
 * Copy the current url to the clipboard.
 */
int
copy_url_to_clipboard(GtkWidget *widget, gpointer data)
{
	GtkClipboard *cb = gtk_clipboard_get(GDK_NONE);
	gtk_clipboard_set_text(cb, (char *)data, strlen((char *)data));
	return 0;
}

/*
 * Add a new string to the statusbar.
 */
int
status_push(char *fmt, ...)
{
	va_list ap;
	char *buf;
	char *buf2 = NULL;
	int len, retval;
	char *site, newlistdata[MAXLN];

	if (!fmt)
		return -1;
	
	va_start(ap, fmt);
	if (!(len = vfsalloc(fmt, ap)))
		return -1;
	va_end(ap);

	if ((buf = malloc(len)) == 0)
		return -1;
	
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	pthread_mutex_lock(&iter_lock);
	pthread_mutex_lock(&store_lock);
	pthread_mutex_lock(&list_lock);
	pthread_mutex_unlock(&list_lock);
	gtk_tree_model_get(GTK_TREE_MODEL(store), &iter_global, 0, &site, -1);
	strlcpy(newlistdata, site, MAXLN);
	buf2 = strstr(site, "(Downloading...");
	if (buf2 != NULL)
		newlistdata[buf2-site] = '\0';
	else
		strlcat(newlistdata, " ", MAXLN);

	strlcat(newlistdata, buf, MAXLN);

	gtk_tree_store_set(store, &iter_global, 0, newlistdata, -1);
	pthread_mutex_unlock(&store_lock);
	pthread_mutex_unlock(&iter_lock);
	free(buf);
	return retval;
}
